import csv
import datetime
import numpy as np
import matplotlib.pyplot as plt

from time import time
from astropy.wcs import WCS
from decimal import Decimal
from .uncertainty import z_uncert
from .flux_zfind import Template, gaussf
from .fits2flux import Fits2flux, wcs2pix, pix2wcs
from .fft_zfind import Fourier, double_damped_sinusoid

# Get the prefix of a unit from an exponent
_prefix = {-24 : 'y', -21 : 'z', -18 : 'a',
        -15 : 'f', -12 : 'p',  -9 : 'n',
        -6 : '\u03BC', -3 : 'm',   0 : '', 
        3 : 'k',   6 : 'M',   9 : 'G',
        12 : 'T',  15 : 'P',  18 : 'E', 
        21 : 'Z',  24 : 'Y'}

def _count_decimals(number: float):
    """ Count the number of decimals in a float """
    d = Decimal(str(number))
    d = abs(d.as_tuple().exponent)
    return d

def _flatten_list(input_list):
    """ Turns lists of lists into a single list """
    flattened_list = []
    for array in input_list:
        for x in array:
            flattened_list.append(x)
    return flattened_list

class zfinder():
    def __init__(self, image, ra, dec, transition, aperture_radius, bvalue):
        """
        zfinder is a wrapper on individual function files fit2flux, Template, and Fourier.
        use zfinder to find the redshift of a fits image via two different methods: zflux
        fits gaussian functions to the data to find redshift; and zfft performs the fast
        fourier transform on the flux data to find redshift. These methods will create and save
        a series of plots and csv files with raw data.

        Parameters
        ----------
        image : .fits
            A .fits image file
        
        ra : list
           Right ascension of the target [h, m, s]
        
        dec : list
            Declination of the target [d, m, s, esign]
        
        aperture_radius : float
            Radius of the aperture to use over the source (pixels)

        bvalue : float
            The value of BMAJ and BMIN (arcseconds)
        """
        self._image = image
        self._ra = ra
        self._dec = dec
        self._transition = transition
        self._aperture_radius = aperture_radius
        self._bvalue = bvalue
        
        self._source = Fits2flux(image, ra, dec, aperture_radius, bvalue)
        self._frequency = self._source.get_freq()
        self._flux, self._uncertainty = self._source.get_flux()
    
    @staticmethod
    def __plot_chi2(z, chi2, dz, title):
        min_chi2 = min(chi2)
        min_z = z[np.argmin(chi2)]
        d = _count_decimals(dz)
        plt.figure(figsize=(20,9))
        plt.plot(z, chi2, color='black')
        plt.plot(min_z, min_chi2, 'bo', markersize=5)
        plt.title(f'{title} $\chi^2_r$ = {round(min_chi2, 2)} @ z={round(min_z, d)}', fontsize=15)
        plt.xlabel('Redshift', fontsize=15)
        plt.ylabel('$\chi^2_r$', x=0.01, fontsize=15)
        plt.yscale('log')
        plt.savefig(f'{title} Chi2.png', dpi=200)
        plt.show()

    @staticmethod
    def __export_csv(title, neg, min_z, pos, params, perr, z, chi2):
        """ Export the data to a csv file """

        # Write chi2 vs redshift and uncertainties
        with open(f'{title}_data.csv', 'w', newline='') as f:
            wr = csv.writer(f)
            
            # Export redshift and uncertainties
            wr.writerow([f'{title} Redshift and uncertainties'])
            wr.writerow(['-', 'z', '+'])
            wr.writerow([neg, min_z, pos])
            wr.writerow([])

            # Export function parameters and uncertainty
            wr.writerow([f'Best Fitting Parameters for the {title} Plot'])
            wr.writerow(['Parameter', 'Value', 'Uncertainty'])
            rows = zip(['amplitude', 'standard deviation'], params, perr)
            for row in rows:
                wr.writerow(row)
            wr.writerow([])

            # Export raw arrays of chi2 for each redshift
            rows = zip(z, chi2)
            wr.writerow(['Redshift', 'Chi-squared'])
            for row in rows:
                wr.writerow(row)
    
    @staticmethod
    def __plot_pixels(snrs, pixels, peaks):
        """ Plot the snr vs # pixels """
        
        # Initialise arrays
        blue_points_x = [pixels[0]]
        blue_points_y = [snrs[0]]
        green_points_x = []
        green_points_y = []
        orange_points_x = []
        orange_points_y = []
        red_points_x = []
        red_points_y = []

        # Sort points
        for snr, pix, pk in zip(snrs[1:], pixels[1:], peaks[1:]):
            num_pks = len(pk)
            
            # If there are no lines
            if len(snr) == 0:
                continue
            
            # bad - no redshift or extreme redshift (z>15)
            if num_pks < 2 or num_pks > 4: 
                green_points_x.append(pix)
                green_points_y.append(snr)
                continue
            
            # Two peaks is a good sign its real
            if num_pks == 2:
                pk_diffs = np.abs(np.diff(pk)) # needs to be at least 200 channels  
                if pk_diffs > 200:
                    red_points_x.append(pix)
                    red_points_y.append(snr)
                else:
                    orange_points_x.append(pix)
                    orange_points_y.append(snr)
            
            # Check 3 or 4 peaks
            else:
                pk_diffs = np.abs(np.diff(pk))            
                pk_diffs = np.abs(np.diff(pk_diffs))
                pk_diffs = np.average(pk_diffs)
                if pk_diffs < 15:
                    red_points_x.append(pix)
                    red_points_y.append(snr)
                else:
                    orange_points_x.append(pix)
                    orange_points_y.append(snr)
        
        # The number of points irrespective of how many sslf lines found
        blue_points = len(blue_points_x)
        green_points = len(green_points_x)
        orange_points = len(orange_points_x)
        red_points = len(red_points_x)

        # Flatten lists of lists to one big list
        blue_points_x = _flatten_list(blue_points_x)
        blue_points_y = _flatten_list(blue_points_y)
        green_points_x = _flatten_list(green_points_x)
        green_points_y = _flatten_list(green_points_y)
        orange_points_x = _flatten_list(orange_points_x)
        orange_points_y = _flatten_list(orange_points_y)
        red_points_x = _flatten_list(red_points_x)
        red_points_y = _flatten_list(red_points_y)
        
        # Random points
        with open(f'snr_vs_pix.csv', 'w', newline='') as f:
            wr = csv.writer(f)
            rows = zip(['Blue', 'Green', 'Yellow', 'Red'], 
                       [blue_points, green_points, orange_points, red_points])
            for row in rows:
                wr.writerow(row)
            wr.writerow([])

        # Make the plot
        plt.figure(figsize=(20,9))
        plt.scatter(blue_points_x, blue_points_y, s=60, marker='*', color='blue')
        plt.scatter(green_points_x, green_points_y, s=60, marker='X', color='green')
        plt.scatter(orange_points_x, orange_points_y, s=60, marker='D', color='darkorange')
        plt.scatter(red_points_x, red_points_y, s=60, marker='s', color='red')
        plt.title(f'No. Random Points = {len(snrs)-1}', fontsize=20)
        plt.xlabel('No. of Pixels', fontsize=20)
        plt.ylabel('SNR', fontsize=20)
        plt.legend(['Target', 'No significance', 'Low significance', 'High significance'])
        plt.xticks(fontsize=15)
        plt.yticks(fontsize=15)
        plt.savefig('SNR vs Pix.png')
        plt.show()
        
    @staticmethod
    def __plot_circle_points(coords, radius, header):
        """ Plot the corodinates of the random points """
        # If there are more than 2 axis, drop them
        w = WCS(header) # Get the world coordinate system
        if header['NAXIS'] > 2:
            w = w.dropaxis(3) # stokes
            w = w.dropaxis(2) # frequency
        
        circle_points = np.transpose(coords)
        points_x = circle_points[0, :] # all x coordinates except the first which is the original
        points_y = circle_points[1, :] # all y coordinates except the first which is the original
        centre_x = points_x[0]
        centre_y = points_y[0]
        circ = plt.Circle((centre_x, centre_y), radius, fill=False, color='blue', label='_nolegend_')
        fig, ax = plt.subplots(subplot_kw={'projection': w})
        fig.set_figwidth(7)
        fig.set_figheight(7)
        ax.add_patch(circ)
        plt.scatter(points_x[0], points_y[0], color='black')
        plt.scatter(points_x[1:], points_y[1:], color='blue')
        plt.title(f'Distribution of {len(coords)-1} spaced random points')
        plt.xlim(-radius-1+centre_x, radius+1+centre_x)
        plt.ylim(-radius-1+centre_y, radius+1+centre_y)
        plt.xlabel('RA', fontsize=15)
        plt.ylabel('DEC', fontsize=15)
        plt.legend(['Target', 'Random'])
        plt.savefig('Point Distribution.png', dpi=200)
        plt.show()
        
    @staticmethod
    def __export_heatmap_csv(delta_z):
        """ Export matrix of delta redshifts"""
        with open(f'heatmap_data.csv', 'w', newline='') as f:
            wr = csv.writer(f)
            for row in delta_z:
                wr.writerow(row)
            wr.writerow([])
        
    def __heatmap(self):
        """ Grid heatmap of the redshift found per pixel """
    
        # Velocity of the gas
        velocities = [3*10**5*((((1+self._best_z)**2 - 1) / ((1+self._best_z)**2 + 1)) - (((1+i)**2 - 1) / ((1+i)**2 + 1))) for i in self._z_fft_pp] # km/s
        self._z_fft_pp = np.array(self._z_fft_pp).tolist()
        d = _count_decimals(self._best_z)

        # Calculate the difference between the redshifts
        all_deltas = []
        for i in range(len(self._z_fft_pp)):
            deltas = []
            for j in range(len(self._z_fft_pp)):
                dat = float(self._z_fft_pp[i][j])
                delta_z = dat - self._best_z
                delta_z = round(delta_z, d)
                deltas.append(delta_z)
            all_deltas.append(deltas)
        
        # Export the redshifts
        self.__export_heatmap_csv(all_deltas)
        
        # If there are more than 2 axis, drop them
        w = WCS(self._source._hdr) # Get the world coordinate system
        if self._source._hdr['NAXIS'] > 2:
            w = w.dropaxis(3) # stokes
            w = w.dropaxis(2) # frequency
        
        # Show a heatmap of the redshifts
        velocities = np.flipud(velocities)
        plt.subplot(projection=w)
        hm = plt.imshow(velocities, cmap='bwr', interpolation='nearest', vmin=-300, vmax=300, 
                extent=[self._x_coords[0], self._x_coords[-1], self._y_coords[0], self._y_coords[-1]], 
                origin='lower')
        plt.colorbar(hm)
        plt.xlabel('RA', fontsize=15)      
        plt.ylabel('DEC', fontsize=15)      
        plt.savefig('FFT PP Heatmap.png')
        plt.show()

    def _z_uncert(self, z, chi2, sigma, reduction):
        """ Caclulate the uncertainty on the best fitting redshift """
        reduced_sigma = sigma**2 / (len(self._flux) - 2*len(self._peaks) - 1)
        neg, pos = z_uncert(z, chi2, reduced_sigma, reduction)
        return neg, pos
    
    def _plot_blind_lines(self):
        """ Helper function to plot sslf lines in the found flux """
        text_offset_high = max(self._flux)/20
        text_offset_low = 0.4*text_offset_high
        
        for i, line in enumerate(self._peaks):
            x = self._frequency[line]
            y = self._flux[line]
            plt.plot(x, y, 'bo')
            plt.text(x, y+text_offset_high, f'snr={self._snrs[i]}', color='blue')
            plt.text(x, y+text_offset_low, f'scale={self._scales[i]}', color='blue')
    
    def __plot_flux(self, params):
        """ Plot the flux with best fitting redshift """
        x_unit, y_unit = self._source.get_exponents()
        d = _count_decimals(self._dz)
        x0 = self._transition/(1+self._min_z)
        plt.figure(figsize=(20,9))
        plt.plot(self._frequency, np.zeros(len(self._frequency)), color='black', linestyle=(0, (5, 5)))
        plt.plot(self._frequency, self._flux, color='black', drawstyle='steps-mid')
        plt.plot(self._frequency, gaussf(self._frequency, *params, x0), color='red')
        self._plot_blind_lines()
        plt.margins(x=0)
        plt.fill_between(self._frequency, self._flux, 0, where=(np.array(self._flux) > 0), color='gold', alpha=0.75)
        plt.title(f'Flux z={round(self._min_z, d)}', fontsize=15)
        plt.xlabel(f'Frequency $({_prefix[x_unit]}Hz)$', fontsize=15)
        plt.ylabel(f'Flux $({_prefix[y_unit]}Jy)$', fontsize=15)
        plt.savefig('Flux Best Fit.png')
        plt.show()
    
    def __plot_fft_flux(self, params):
        """ Plot the FFT with best fitting redshift"""
        d = _count_decimals(self._dz)
        plt.figure(figsize=(20,9))
        plt.plot(self._ffreq, self._fflux, color='black', drawstyle='steps-mid')
        plt.plot(self._ffreq, np.zeros(len(self._fflux)), color='black', linestyle=(0, (5, 5)))
        plt.plot(self._ffreq, double_damped_sinusoid(self._ffreq, *params, 
            self._dz, self._frequency[0], self._transition), color='red')
        plt.title(f'FFT z={round(self._min_z, d)}', fontsize=15)
        plt.xlabel('Scale', fontsize=15)
        plt.ylabel('Amplitude', fontsize=15)
        plt.margins(x=0)
        plt.savefig('FFT Best Fit.png', dpi=200)
        plt.show()
    
    def zflux(self, z_start=0, dz=0.01, z_end=10, sigma=1, penalise=True, reduction=False):
        """ 
        Finds the best redshift by fitting gaussian functions overlayed on flux data. The
        chi-squared is calculated at every redshift by iterating through delta-z. The most 
        likely redshift corresponds to the minimum chi-squared. Saves and shows the chi2 
        vs z plot, flux plot, and csv file with additional information
        
        Parameters
        ----------
        z_start : int, optional
            The first value in the redshift list. Default = 0
        
        dz : float, optional
            The change in redshift. Default = 0.01
        
        z_end : int, optional
            The final value in the redshift list. Default = 10
        
        sigma : float, optional
            The significance level of the uncertainty in the redshift 
            found at the minimum chi-squared. Default = 1
            
        penalise : bool, optional
            If True, perform chi-squared penalisation with sslf. Default = True
        
        reduction : bool, optional
            Can greatly reduce uncertainty. Choose to reduce all chi2 values
            such that the minimum chi2 is 0. Default = False
        """
        self._dz = dz
        
        # Find the redshift 
        self._flux_source = Template(self._transition, self._frequency, self._flux, self._uncertainty)
        z, chi2 = self._flux_source.zfind(z_start, dz, z_end, sigma, penalise)
        if reduction:
            min_chi2 = min(chi2)
            chi2 = [i - min_chi2 for i in chi2]
        params, perr = self._flux_source.parameters()
        self._peaks, self._snrs, self._scales = self._flux_source.sslf()
        self._min_z = z[np.argmin(chi2)]
        
        # Save the data
        self.__plot_chi2(z, chi2, dz, 'Flux')
        self.__plot_flux(params)
        neg, pos = self._z_uncert(z, chi2, sigma, reduction=reduction)
        self.__export_csv('flux', neg, self._min_z, pos, params, perr, z, chi2)
        
    def zfft(self, z_start=0, dz=0.01, z_end=10, sigma=1, reduction=False):
        """ 
        Finds the best redshift by performing the fast fourier transform on the flux data. The
        chi-squared is caclulated at every redshift by iterating through delta-z. The most 
        likely redshift corresponds to the minimum chi-squared. Saves and shows the chi2 vs z 
        plot, flux plot, and csv file with additional information.
        
        Parameters
        ----------        
        z_start : int, optional
            The first value in the redshift list. Default = 0
            
        dz : float, optional
            The change in redshift. Default = 0.01
        
        z_end : int, optional
            The final value in the redshift list. Default = 10
        
        sigma : float, optional
            The significance level of the uncertainty in the redshift 
            found at the minimum chi-squared. Default = 1
        
        reduction : bool, optional
            Can greatly reduce uncertainty. Choose to reduce all chi2 values
            such that the minimum chi2 is 0. Default = False
        """
        
        self._dz = dz
        
        # Get the fft flux data
        self._ffreq, self._fflux = Fourier.fft(self._frequency, self._flux)
        
        # Get the redshift
        fft_source = Fourier(self._transition, self._frequency, self._flux)
        z, chi2 = fft_source.zfind(z_start, dz, z_end, sigma)
        if reduction:
            min_chi2 = min(chi2)
            chi2 = [i - min_chi2 for i in chi2]
        params, perr = fft_source.fft_params()
        self._all_num_peaks = fft_source._peaks
        self._min_z = z[np.argmin(chi2)]
        
        # Save the data
        self.__plot_chi2(z, chi2, dz, 'FFT')
        self.__plot_fft_flux(params)
        neg, pos = self._z_uncert(z, chi2, sigma, reduction=reduction)
        self.__export_csv('FFT', neg, self._min_z, pos, params, perr, z, chi2)
    
    def random_stats(self, n=100, radius=50, spread=1):
        """ 
        Iterate through n-1 randomly generated coordinates to find the 
        signal-to-noise ratio, number of pixels, and channel peaks for each.
        The first position is always the target: if n=100, only 99 are
        random, with the first being the target.
        
        Parameters
        ----------
        n : int, optional
            The number of radnom points to find statistics for. Default = 100
        
        radius : float, optional
            The radius of the image to find statistics (in pixels) centred on
            the given ra and dec. Default = 50
        
        spread : float, optional
            The minimum spread of random points (in pixels). Default = 1
        
        Returns
        -------
        snrs : list
            A list of signifcant point signal-to-noise ratios
            
        pixels : list
            The list of pixels for each random position
        
        peaks : list
            The channel location of all peaks in every position
        """
        
        snrs, pixels, peaks = self._source.random_analysis(n, radius, spread)
        self.__plot_circle_points(self._source._coordinates, radius, self._source._hdr)        
        self.__plot_pixels(snrs, pixels, peaks)
    
    def fft_per_pixel(self, size, z_start=0, dz=0.01, z_end=10, sigma=1):
        """ 
        Performs the FFT redshift finding method in a square around the 
        target right ascension and declination. Automatically saves and
        shows a heatmap of the distribution of the redshifts.
        
        Parameters
        ----------
        size : int
            size to perform the fft method around the centre. size=3 performs
            a 3x3, size=5 a 5x5, etc. 
                
        z_start : int, optional
            The first value in the redshift list. Default = 0
            
        dz : float, optional
            The change in redshift. Default = 0.01
        
        z_end : int, optional
            The final value in the redshift list. Default = 10
        
        sigma : float, optional
            The significance level of the uncertainty in the redshift 
            found at the minimum chi-squared. Default = 1
        """
        
        # Get the redshift
        fft_source = Fourier(self._transition, self._frequency, self._flux)
        z, chi2 = fft_source.zfind(z_start, dz, z_end, sigma)
        self._best_z = z[np.argmin(chi2)]
        self._centre_coords = wcs2pix(self._ra, self._dec, self._source._hdr)
        
        # If size is even, make it odd
        if size %2 == 0:
            size += 1
        
        # Generate the x and y coordinates to iterate through
        matrix = np.zeros(size)
        for i in range(len(matrix)):
            matrix[i] = i-len(matrix)//2
        self._x_coords = matrix + self._centre_coords[0]
        self._y_coords = matrix + self._centre_coords[1]
        
        i = 0
        z_fft_pp = []
        d = _count_decimals(dz)
        for y in reversed(self._y_coords):
            for x in self._x_coords:
                start = time()
        
                ra, dec = pix2wcs(x, y, self._source._hdr)
                
                gleam_0856 = Fits2flux(self._image, ra, dec, self._aperture_radius, self._bvalue)
                freq = gleam_0856.get_freq()
                flux, uncert = gleam_0856.get_flux()

                zf = Fourier(self._transition, freq, flux)
                z, chi2 = zf.zfind(z_start, dz, z_end)

                lowest_z = z[np.argmin(chi2)]
                lowest_z = round(lowest_z, d)

                z_fft_pp.append(lowest_z)

                end = time()
                elapsed = end - start
                
                remaining = datetime.timedelta(seconds=round((size**2-(i+1))*elapsed))
                
                print(f'{i+1}/{size**2}, took {round(elapsed,2)} seconds, approx {remaining} remaining')
                i+=1

        self._z_fft_pp = np.array_split(z_fft_pp, size)
        self.__heatmap()
        return z_fft_pp


import numpy as np
from .flux_zfind import gaussf
from scipy.signal import find_peaks
from scipy.optimize import curve_fit

def double_damped_sinusoid(x, a, s, z, nu, f):
    """ FFT Fitting function """
    A = 22.09797605*a*s
    C = 1/(np.pi*s)
    N = np.floor(((1+z)*nu/f)+1)
    p = 2*np.pi*(N*f/(1+z) - nu)
    q = 2*np.pi*((N+1)*f/(1+z) - nu)
    y = A*np.exp(-(x / C)**2) * (np.cos(p*x) + np.cos(q*x))
    return y

def _find_params(transition, ffreq, fflux, dz, x0):
    """ Find the best fitting parameters for the FFT """
    try:
        params, covars = curve_fit(lambda x, a, s: double_damped_sinusoid(x, a, s, z=dz, 
            nu=x0, f=transition), ffreq, fflux, bounds=[[0, 0], [max(fflux), 2]])
    except:
        return np.array([None, None]), np.array([None, None])
    return params, covars

class Fourier():
    def __init__(self, transition, frequency, flux):
        """
        Find statistics of the source via fast fourier transform
        
        Parameters
        ----------
        transition : float
            The first transition frequency of the element or molecule to search for. Units
            follow frequency
        
        frequency : list
            A list of frequency values calculated from fits2flux
        
        flux : list
            A list of flux values calculated from fits2flux
        """
        self._transition = transition
        self._frequency = frequency
        self._flux = flux
    
    @staticmethod
    def fft(frequency, flux):
        """ 
        Performs the fast fourier transform on the frequency and flux axis 
        
        Parameters
        ----------
        frequency : list
            Frequency axis values found with fits2flux `get_freq()`
        
        flux : list
            Flux axis values found with fit2flux `get_flux()`
        
        Returns
        -------
        ffreq, fflux : list
            The Fourier Transformed frequency and flux axes respectively
        """
        N = 10*len(flux) # Number of sample points
        T = frequency[1]-frequency[0] # sample spacing

        # Fourier transformed data
        fflux = np.fft.rfft(flux, N).real
        ffreq = np.fft.rfftfreq(N, T)
        return ffreq, fflux
    
    @staticmethod
    def __calc_all_num_gauss(transition, frequency, dz):
        """ Calculate the number of gaussians inside the window (x-axis) """
        loc = transition/(1+dz)
        f_exp = gaussf(frequency, a=0.5, s=0.5, x0=loc)

        # Caclulate the number of gaussians overlayed
        gauss_peaks = find_peaks(f_exp)[0]
        return gauss_peaks

    def zfind(self, z_start=0, dz=0.01, z_end=10, sigma=1):
        """ 
        Finds the best redshift by performing the fast fourier transform on the flux data. The
        chi-squared is caclulated at every redshift by iterating through delta-z. The most 
        likely redshift corresponds to the minimum chi-squared.
        
        Parameters
        ----------        
        z_start : int, optional
            The first value in the redshift list. Default = 0
            
        dz : float, optional
            The change in redshift. Default = 0.01
        
        z_end : int, optional
            The final value in the redshift list. Default = 10
        
        sigma : float, optional
            The significance level of the uncertainty in the redshift 
            found at the minimum chi-squared. Default = 1
        
        Returns
        -------
        z : list
            The list of redshifts that was used to calculate the chi-squared
        
        chi2 : list
            A list of calculated chi-squared values
        """
        
        # Initialise lists
        z = np.arange(z_start, z_end+dz, dz)
        sigma = sigma
        all_chi2 = []
        fft_params = []
        fft_perrs = []
        all_num_peaks = []
        
        # Fourier transform frequency and flux
        ffreq, fflux = self.fft(self._frequency, self._flux)
        x0 = self._frequency[0]

        # Interate through the list of redshifts and calculate the chi-squared
        for dz in z:
            
            # Find the best fitting parameters at this redshift
            params, covars = _find_params(self._transition, ffreq, fflux, dz, x0)
            if not params.tolist()[0]:
                all_chi2.append(max(all_chi2))
                fft_params.append([99,99])
                fft_perrs.append([99,99])
                all_num_peaks.append(0)
                continue
            perr = np.sqrt(np.diag(covars))

            # Calulate chi-squared
            fflux_obs = double_damped_sinusoid(ffreq, *params, z=dz, nu=x0, f=self._transition)

            # Find the number of gaussians that would be overlayed at this redshift
            gauss_peaks = self.__calc_all_num_gauss(self._transition, self._frequency, dz)
            num_gauss_peaks = len(gauss_peaks)

            chi2 = sum((fflux - fflux_obs)**2) # chi-squared
            reduced_chi2 = chi2 / (len(fflux_obs) - 2*num_gauss_peaks - 1)

            # Append all items
            all_chi2.append(reduced_chi2)
            fft_params.append(params)
            fft_perrs.append(perr)
            all_num_peaks.append(num_gauss_peaks)
        
        lowest_index = np.argmin(all_chi2)
        self._params = fft_params[lowest_index]
        self._perr = fft_perrs[lowest_index]
        self._peaks = all_num_peaks[lowest_index]
        
        return z, all_chi2
    
    def fft_params(self):
        """
        Get the paramters of the best fitting redshift. Must run fft_zfind first.
        
        Returns
        -------
        params : list
            Best fitting redshift parameters -> [amplitude, standard deviation]
        
        perr : list
            Errors on the best fitting redshift parameters -> [amplitude, standard deviation] 
        """
        return self._params, self._perr


import datetime
import numpy as np

from time import time
from random import random
from astropy.io import fits
from astropy.wcs import WCS
from .flux_zfind import find_lines
from warnings import filterwarnings
from scipy.spatial.distance import cdist
from photutils.background import Background2D
from PyAstronomy.pyasl import hmsToDeg, dmsToDeg, degToHMS, degToDMS
from photutils.aperture import CircularAperture, CircularAnnulus, ApertureStats, aperture_photometry

def _drop_axis(hdr):
    """ Drop extra axes """
    w = WCS(hdr) # Get the world coordinate system
    if hdr['NAXIS'] > 2:
        w = w.dropaxis(3) # stokes
        w = w.dropaxis(2) # frequency
    return w

def wcs2pix(ra, dec, hdr):
    """ Convert ra and dec to x, y positional pixel world coordinates """
    
    # Drop extra axes
    w = _drop_axis(hdr)

    # Convert to decimal degrees
    ra = hmsToDeg(ra[0], ra[1], ra[2])
    dec = dmsToDeg(dec[0], dec[1], dec[2], esign=dec[3])

    # Convert world coordinates to pixel
    x, y = w.all_world2pix(ra, dec, 1)

    # Round to nearest integer
    x = int(np.round(x))
    y = int(np.round(y))
    return [x, y]

def pix2wcs(x, y, hdr):
    """ Convert x, y positional pixel coordinates to world ra and dec """
    
    # Drop extra axes
    w = _drop_axis(hdr)

    # Convert x,y to degrees
    ra, dec = w.all_pix2world(x, y, 1)

    # Convert degrees to ra, dec
    ra = degToHMS(ra)
    dec = degToDMS(dec)
    return ra, dec

def get_sci_exponent(number):
    """ Find the scientific exponent of a number """
    abs_num = np.abs(number)
    base = np.log10(abs_num) # Log rules to find exponent
    exponent = int(np.floor(base)) # convert to floor integer
    return exponent

def get_eng_exponent(number):
    """ 
    Find the nearest power of 3 (lower). In engineering format,
    exponents are multiples of 3.
    """
    exponent = get_sci_exponent(number) # Get scientific exponent
    for i in range(3):
        
        if exponent > 0:
            unit = exponent-i
        else:
            unit = exponent+i
            
        if unit % 3 == 0: # If multiple of 3, return it
            return unit

class Fits2flux(object):
    def __init__(self, image, ra, dec, aperture_radius, bvalue):
        """ 
        Calculate the flux over the frequency range of a .fits image
        
        Parameters
        ----------
        image : .fits
            A .fits image file
        
        ra : list
           Right ascension of the target [h, m, s]
        
        dec : list
            Declination of the target [d, m, s, esign]
        
        aperture_radius : float
            Radius of the aperture to use over the source (pixels)

        bvalue : float
            The value of BMAJ and BMIN (arcseconds)
        """
        image = fits.open(image)
        self._hdr = image[0].header
        self._data = image[0].data[0]
        self._ra = ra
        self._dec = dec
        self._aperture_radius = aperture_radius
        self._bvalue = bvalue

        # Ignore warnings (there are hundreds)
        filterwarnings("ignore", module='photutils.background')
        filterwarnings("ignore", module='astropy.wcs.wcs')
    
    @staticmethod
    def __average_zeroes(array):
        """ Average zeroes with left & right values in a list """
        for i, val in enumerate(array):
            if val == 0:
                try:
                    array[i] = (array[i-1] + array[i+1])/2
                except IndexError:
                    array[i] = (array[i-2] + array[i-1])/2
        return array

    @staticmethod
    def __calc_beam_area(bvalue):
        """ Caclulate the corrected beam area (from Jy/beam to Jy) """
        bmaj = bvalue/3600
        bmin = bmaj
        barea = 1.1331 * bmaj * bmin
        return barea

    @staticmethod
    def __calc_pixels(spec_peaks, flux):
        """ 
        Calculate the number of pixels above 0 Jy left & right of all lines found 
        by the sslf line finder.
        """
        pixels = []
        for peak in spec_peaks:
            num_pix = 0

            # left side
            i = 1
            try:
                while flux[peak-i] > 0:
                    num_pix += 1
                    i += 1
            except IndexError:
                pass

            # right side
            i = 0
            try:
                while flux[peak+i] > 0:
                    num_pix += 1
                    i += 1
            except IndexError:
                pass

            pixels.append(num_pix)
        return pixels
    
    @staticmethod
    def get_coords(n, radius, centre, min_spread):
        """ 
        Calculate random coordinates in a circle with a minimum spread 
        
        Parameters
        ----------
        n : int
            Number of random coordinates (minus 1) to generate.
            
        radius : float
            The radius of the circle (pixels).
            
        centre : list, optional
            [x, y] positional centre of the circle. Default=[0,0]
            
        min_spread : float
            Minimum distance between all coordinates. Default=1
            
        Returns
        -------
        coordinates : list
            List of coordinates; [[x1,y1], [x2,y2], ... [xn,yn]]        
        """
        coords = [centre] # centre = [x,y] -> coords = [[x,y]]
        
        # Iterate through the number of coords.
        for _ in range(n-1):
            
            # Keep generating the current point until it is at least the minimum distance away from all 
            while True:
                theta = 2 * np.pi * random() # choose a random direction
                r = radius * random() # choose a random radius

                # Convert coordinates to cartesian
                x = r * np.cos(theta) + centre[0]
                y = r * np.sin(theta) + centre[1]

                # Find the distance between all the placed coords
                distances = cdist([[x,y]], coords, 'euclidean')
                min_distance = min(distances[0])
                
                # If the minimum distance is satisfied for all coords, go to next point
                if min_distance >= min_spread or len(coords) == 1:
                    coords.append([x,y])
                    break
        return coords

    def get_freq(self):
        """ 
        Caclulate the frequency axis (x-axis) of the flux

        Returns
        -------
        frequency : list
            A list of frequencies corresponding to individual channels of a .fits image
        """
        
        # Required values
        start = self._hdr['CRVAL3']
        increment = self._hdr['CDELT3']
        length = self._hdr['NAXIS3']

        # Get the exponent
        self._xexponent = get_eng_exponent(start)

        # Normalise start and increment to the exponent
        norm_factor = 10**self._xexponent
        start /= norm_factor
        increment /= norm_factor

        # Calculate end point
        end = start + length * increment

        # Create axis list
        frequency = np.linspace(start, end, length)
        return frequency

    def get_flux(self):
        """ 
        For every frequency channel, find the flux and associated uncertainty at a position
        
        Returns
        -------
        flux : list
            A list of flux values from each frequency channel
        
        f_uncert : list
            A list of flux uncertainty values for each flux measurement        
        """

        # Initialise array of fluxes and uncertainties to be returned
        flux = []
        f_uncert = []
        
        # Calculate area of the beam
        barea = self.__calc_beam_area(self._bvalue)
        pix2deg = self._hdr['CDELT2'] # Pixel to degree conversion factor

        # The position to find the flux at
        position = wcs2pix(self._ra, self._dec, self._hdr)
        
        # For every page of the 3D data matrix, find the flux around a point (aperture)
        for channel in self._data:

            # Setup the apertures 
            aperture = CircularAperture(position, self._aperture_radius)
            annulus = CircularAnnulus(position, r_in=2*self._aperture_radius, r_out=3*self._aperture_radius)

            # Uncertainty
            aperstats = ApertureStats(channel, annulus) 
            rms  = aperstats.std 

            # Background
            bkg = Background2D(channel, (50, 50)).background 

            # Aperture sum of the fits image minus the background
            apphot = aperture_photometry(channel - bkg, aperture)
            apsum = apphot['aperture_sum'][0]

            # Calculate corrected flux
            total_flux = apsum*(pix2deg**2)/barea

            # Save values
            flux.append(total_flux)
            f_uncert.append(rms)

        # Average 0's from values left & right
        f_uncert = self.__average_zeroes(f_uncert) 
        
        # Normalising factor
        self._yexponent = get_eng_exponent(np.max(flux))
        norm_factor = 10**-self._yexponent

        flux = np.array(flux)*norm_factor
        f_uncert = np.array(f_uncert)*norm_factor
        return flux, f_uncert
    
    def random_analysis(self, n=100, radius=50, min_spread=1):
        """ 
        Iterate through n randomly generated coordinates to find the 
        signal-to-noise ratio, number of pixels, and channel peaks for each.
        
        Parameters
        ----------
        n : int, optional
            The number of points to find statistics for. Default = 100
        
        radius : float, optional
            The radius of the image to find statistics (in pixels) centred on
            the given ra and dec. Default = 50
        
        min_spread : float, optional
            The minimum spread of random points (in pixels). Default = 1
        
        Returns
        -------
        snrs : list
            A list of signifcant point signal-to-noise ratios
            
        pixels : list
            The list of pixels for each random position
        
        peaks : list
            The channel location of all peaks in every position
        """
        
        # Initialise return arrays
        all_snrs = []
        all_pixels = []
        all_spec_peaks = []

        # Find x,y coordinates of the target
        x, y = wcs2pix(self._ra, self._dec, self._hdr)

        # Calculate the coordinates for the number of points
        self._coordinates = self.get_coords(n, radius, [x,y], min_spread)

        # Use sslf to find lines on all points
        for i, (x, y) in enumerate(self._coordinates):
            start = time()
            self._ra, self._dec = pix2wcs(x,y, self._hdr) # convert x, y to ra,dec

            flux, uncert = self.get_flux()
            
            spec_peaks, snrs, scales = find_lines(flux) # use flux axis to find lines on
            
            pixels = self.__calc_pixels(spec_peaks, flux)

            all_snrs.append(snrs)
            all_pixels.append(pixels)
            all_spec_peaks.append(spec_peaks)
            
            end = time()
            elapsed = end - start 
            remaining = datetime.timedelta(seconds=round(elapsed*(n-(i+1))))
            
            print(f'{i+1}/{len(self._coordinates)}, took {round(elapsed,2)} seconds, approx {remaining} remaining')

        return all_snrs, all_pixels, all_spec_peaks
    
    def get_exponents(self):
        """ get the x and y factors that were used to normalise the data. 
            -9 = nano, -6 = micro, -3 = milli, 3 = kilo, 6 = mega, 9 = giga,
            etc, etc, etc.
        
        Returns
        -------
        x, y : int
            The x and y axes normalisation factors
        """
        return self._xexponent, self._yexponent


from numpy import sqrt, argmin
from scipy.optimize import curve_fit

def quadratic(x, a, b, c):
    """ Fit a quadratic to 3 points closest to the minimum chi-squared """
    y = a*(x-b)**2 + c
    return y

def solve_quadratic(y, a, b, c):
    """ Solve the quadratic to find x at a certain sigma """
    neg = -sqrt((y-c)/a) + b
    pos = sqrt((y-c)/a) + b
    return neg, pos

def z_uncert(z, chi2, ssigma, reduction=False):
    """
    Caclulate the uncertainty in the best fitting redshift

    Parameters
    ----------
    z : list
        A list of redshift values
    
    chi2 : list
        A list of chi-squared values

    ssigma : float
        The squared significance level of the uncertainty. ssigma = sigma**2
    
    reduction : bool, optional
        Choose to reduce all chi2 values such that the minimum chi2 is 0. Default = False
    
    Returns
    -------
    neg_uncert : float
        The left (- / negative) uncertainty

    pos_uncert : float
        The right (+ / positive) uncertainty
    """
    
    lowest_y_index = argmin(chi2)
    min_x = z[lowest_y_index]
    min_y = min(chi2)
    
    if reduction:
        chi2 = [i - min_y for i in chi2]

    # Get points left and right of lowest_y that is above lowest_y + sigma
    left = lowest_y_index-1
    right = lowest_y_index+1

    # Create the axis arrays for the line between the left, lowest, and right point
    points_x = [z[left], min_x, z[right]]
    points_y = [chi2[left], min_y, chi2[right]]
    
    # Find the parameters of the parabola that connects left, lowest, & right
    params, covars = curve_fit(lambda x, a: quadratic(x, a, b=min_x, c=min_y), points_x, points_y)

    # Calculate the x points above lowest_y + sigma 
    neg, pos = solve_quadratic(min_y + ssigma, *params, b=min_x, c=min_y)

    # Calculate the uncertainty in x
    neg_uncert = min_x - neg
    pos_uncert = pos - min_x
    return neg_uncert, pos_uncert